# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from copy import copy, deepcopy
from itertools import izip

SENTINEL = object()

SHIFT = 5
BMAP = (1 << SHIFT) - 1
BRANCH = 2 ** SHIFT

MAXBITMAPDISPATCH = 16

def relevant(hsh, shift):
    """ Return the relevant part of the hsh on the level shift. """
    return hsh >> shift & BMAP


POPCOUNT_TBL = [0] * (2 ** 16)
for idx in xrange(2 ** 16):
    POPCOUNT_TBL[idx] = (idx & 1) + POPCOUNT_TBL[idx >> 1]

def bit_count(v):
    return (POPCOUNT_TBL[v & 0xffff] +
            POPCOUNT_TBL[(v >> 16) & 0xffff])


def doc(docstring):
    """ Decorator to set docstring of function to docstring. """
    def deco(fn):
        """ Implementation detail. """
        fn.__doc__ = docstring
        return fn
    return deco


ASSOC = "\n".join([
    "Add AssocNode node whose key's hash is hsh to the node or its children.",
    "shift refers to the current level in the tree, which must be a multiple",
    "of the global constant BRANCH. If a node with the same key already",
    "exists, override it.",
])

IASSOC = "\n".join([
    "Modify so that the AssocNode whose key's hash is hsh is added to it.",
    "USE WITH CAUTION.",
    "shift refers to the current level in the tree, which must be a multiple",
    "of the global constant BRANCH. If a node with the same key already",
    "exists, override it.",
])

GET = "\n".join([
    "Get value of the AssocNode with key whose hash is hsh in the subtree.",
    "shift refers to the current level in the tree, which must be a multiple",
    "of the global constant BRANCH.",
])

WITHOUT = "\n".join([
    "Remove AssocNode with key whose hash is hsh from the subtree.",
    "shift refers to the current level in the tree, which must be a multiple",
    "of the global constant BRANCH.",
])

IWITHOUT = "\n".join([
    "Modify so that the AssocNode whose key's hash is hsh is removed from it.",
    "USE WITH CAUTION.",
    "shift refers to the current level in the tree, which must be a multiple",
    "of the global constant BRANCH.",
])


class Node(object):
    __slots__ = []
    def __and__(self, other):
        new = NULLNODE
        
        for node in other:
            try:
                self.get(hash(node.key), 0, node.key)
            except KeyError:
                pass
            else:
                new = new._iassoc(hash(node.key), 0, node)
        return new
    
    def __xor__(self, other):
        new = self
        for node in other:
            new = new.xor(node.hsh, 0, node)
        return new
    
    def __or__(self, other):
        new = self
        for node in other:
            new = new.assoc(node.hsh, 0, node)
        return new
    
    def __eq__(self, other):
        return all(node == othernode for node, othernode in izip(self, other))
    
    def __neq__(self, other):
        return any(node != othernode for node, othernode in izip(self, other))


class NullNode(Node):
    """ Dummy node being the leaf of branches that have no entries. """
    __slots__ = []
    def xor(self, hsh, shift, node):
        return node
    
    _ixor = xor
    
    @doc(ASSOC)
    def assoc(self, hsh, shift, node):
        # Because there currently no node, the new node
        # is the node to be added.
        return node
    
    # The NullNode does not need to be modified if a new association is
    # created because it only returns the new node, hence _iassoc = assoc.
    _iassoc = assoc
    
    def get(self, hsh, shift, key):
        # There is no entry with the searched key because the hash leads
        # to a branch ending in a NullNode.
        raise KeyError(key)
    
    @doc(WITHOUT)
    def without(self, hsh, shift, key):
        # There is no entry with the key to be removed because the hash leads
        # to a branch ending in a NullNode.
        raise KeyError(key)
    
    _iwithout = without
    
    def __iter__(self):
        # There are no keys contained in a NullNode. Hence, an empty
        # iterator is returned.
        return iter([])
    
    # Likewise, there are no values and items in a NullNode.
    iteritems = itervalues = __iter__
    
    def __copy__(self):
        return self
    
    def cutoff(self, hsh):
        return self


# We only need one instance of a NullNode because it does not contain
# any data.
NULLNODE = NullNode()


class HashCollisionNode(Node):
    """ If hashes of two keys collide, store them in a list and when a key
    is searched, iterate over that list and find the appropriate key. """
    __slots__ = ['children', 'hsh']
    def __init__(self, nodes):
        self.children = nodes
        self.hsh = hash(nodes[0].hsh)

    def xor(self, hsh, shift, node):
        if not any(node.key == child.key for child in self.children):
            return HashCollisionNode(self.children + [node])
        return self
    
    def _ixor(self, hsh, shift, node):
        if not any(node.key == child.key for child in self.children):
            self.children.append(node)
        return self
    
    @doc(GET)
    def get(self, hsh, shift, key):
        # To get the child we want we need to iterate over all possible ones.
        # The contents of children are always AssocNodes,
        # so we can safely access the key member.
        for node in self.children:
            if key == node.key:
                return node
        raise KeyError(key)
    
    @doc(ASSOC)
    def assoc(self, hsh, shift, node):
        # If we have yet another key with a colliding key, return a new node
        # with it added to the children, otherwise return a DispatchNode.
        if hsh == self.hsh:
            return HashCollisionNode(self.children + [node])
        return DispatchNode.make(shift, [self, node])
    
    @doc(IASSOC)
    def _iassoc(self, hsh, shift, node):
        # If we have yet another key with a colliding key, add it to the
        # children, otherwise return a DispatchNode.
        if hsh == self.hsh:
            self.children.append(node)
            return self
        return DispatchNode.make(shift, [self, node])
    
    @doc(WITHOUT)
    def without(self, hsh, shift, key):
        # Remove the node whose key is key from the children. If it was the
        # last child, return NULLNODE. If there was no member with a
        # matching key, raise KeyError.
        newchildren = [node for node in self.children if node.key != key]
        if not newchildren:
            return NULLNODE
        
        if newchildren == self.children:
            raise KeyError(key)
        
        return HashCollisionNode(newchildren)
    
    @doc(IWITHOUT)
    def _iwithout(self, hsh, shift, key):
        newchildren = [node for node in self.children if node.key != key]
        if not newchildren:
            return NULLNODE
        
        if newchildren == self.children:
            raise KeyError(key)
        
        self.children = newchildren
        return self
    
    def __iter__(self):
        for node in self.children:
            for elem in node:
                yield elem
    
    def __copy__(self):
        return HashCollisionNode(map(copy, self.children))
    
    def cutoff(self, hsh):
        if self.hsh <= hsh:
            return NULLNODE
        return self


class ListDispatch(Node):
    """ Light weight dictionary like object for a little amount of items.
    Only feasable for a little amount of items as a list of length nitems 
    is always stored.
    
    Only accepts integers as keys. """
    __slots__ = ['items']
    
    def __init__(self, nitems=None, items=None):
        if items is None:
            items = [SENTINEL for _ in xrange(nitems)]
        self.items = items
    
    def replace(self, key, item):
        """ Return a new ListDispatch with the the keyth item replaced
        with item. """
        return ListDispatch(
            None,
            self.items[:key] +
            [item] +
            self.items[key + 1:]
        )
    
    def _ireplace(self, key, item):
        """ Replace keyth item with item.
        
        USE WITH CAUTION. """
        self.items[key] = item
        return self
    
    def __getitem__(self, key):
        value = self.items[key]
        if value is SENTINEL:
            raise KeyError(key)
        return value
    
    def get(self, key, default):
        """ Get keyth item. If it is not present, return default. """
        value = self.items[key]
        if value is not SENTINEL:
            return value
        return default
    
    def remove(self, key):
        """ Return new ListDispatch with keyth item removed.
        Will not raise KeyError if it was not present. """        
        return self.replace(key, SENTINEL)

    def _iremove(self, key):
        """ Remove keyth item. Will not raise KeyError if it was not present.
        
        USE WITH CAUTION. """
        self._ireplace(key, SENTINEL)
        return self
    
    def to_bitmapdispatch(self):
        dispatch = BitMapDispatch()
        for key, value in enumerate(self.items):
            if value is not SENTINEL:
                dispatch._ireplace(key, value)
        return dispatch
    
    def __iter__(self):
        return (item for item in self.items if item is not SENTINEL)
    
    def __copy__(self):
        return ListDispatch(items=self.items[:])
    
    def __deepcopy__(self):
        return ListDispatch(items=map(deepcopy, self.items))
    
    def map(self, fn):
        return ListDispatch(
            items=[SENTINEL if elem is SENTINEL else fn(elem)
                   for elem in self.items]
        )


class BitMapDispatch(Node):
    """ Light weight dictionary like object for a little amount of items.
    Best used for as most as many items as an integer has bits (usually 32).
    Only accepts integers as keys.
    
    The items are stored in a list and whenever an item is added, the bitmap
    is ORed with (1 << key) so that the keyth bit is set.
    The amount of set bits before the nth bit is used to find the index of the
    item referred to by key in the items list.
    """
    __slots__ = ['bitmap', 'items']
    def __init__(self, bitmap=0, items=None):
        if items is None:
            items = []
        self.bitmap = bitmap
        self.items = items
    
    def replace(self, key, item):
        """ Return a new BitMapDispatch with the the keyth item replaced
        with item. """
        # If the item already existed in the list, we need to replace it.
        # Otherwise, it will be added to the list at the appropriate
        # position.
        if len(self.items) >= MAXBITMAPDISPATCH:
            new = self.to_listdispatch(BRANCH)
            return new._ireplace(key, item)
        
        notnew = bool(self.bitmap & 1 << key)
        newmap = self.bitmap | 1 << key
        idx = bit_count(self.bitmap & ((1 << key) - 1))
        return BitMapDispatch(
            newmap,
            # If notnew is True, the item that is replaced by the new item
            # is left out, otherwise the new item is inserted. Refer to
            # _ireplace for a more concise explanation.
            self.items[:idx] + [item] + self.items[idx+notnew:]
        )
    
    def _ireplace(self, key, item):
        """ Replace keyth item with item.
        
        USE WITH CAUTION. """
        if len(self.items) >= MAXBITMAPDISPATCH:
            new = self.to_listdispatch(BRANCH)
            return new._ireplace(key, item)
        
        notnew = bool(self.bitmap & 1 << key)
        self.bitmap |= 1 << key
        idx = bit_count(self.bitmap & ((1 << key) - 1))
        if idx == len(self.items):
            self.items.append(item)
        elif notnew:
            self.items[idx] = item
        else:
            self.items.insert(idx, item)
        
        return self
    
    def get(self, key, default=None):
        """ Get keyth item. If it is not present, return default. """
        if not self.bitmap & 1 << key:
            return default
        return self.items[bit_count(self.bitmap & ((1 << key) - 1))]
    
    def remove(self, key):
        """ Return new BitMapDispatch with keyth item removed.
        Will not raise KeyError if it was not present. """
        idx = bit_count(self.bitmap & ((1 << key) - 1))
        return BitMapDispatch(
            # Unset the keyth bit.
            self.bitmap & ~(1 << key),
            # Leave out the idxth item.
            self.items[:idx] + self.items[idx+1:]
        )
    
    def _iremove(self, key):
        """ Remove keyth item. Will not raise KeyError if it was not present.
        
        USE WITH CAUTION. """
        idx = bit_count(self.bitmap & ((1 << key) - 1))
        self.bitmap &= ~(1 << key)
        self.items.pop(idx)
        return self
    
    def __getitem__(self, key):
        if not self.bitmap & 1 << key:
            raise KeyError(key)
        return self.items[bit_count(self.bitmap & ((1 << key) - 1))]
    
    def to_listdispatch(self, nitems):
        """ Return ListDispatch with the same key to value connections as this
        BitMapDispatch. """
        return ListDispatch(
            None, [self.get(n, SENTINEL) for n in xrange(nitems)]
        )
    
    def __iter__(self):
        return iter(self.items)
    
    def __nonzero__(self):
        return bool(self.items)
    
    def __copy__(self):
        return BitMapDispatch(self.bitmap, self.items[:])
    
    def __deepcopy__(self):
        return BitMapDispatch(self.bitmap, map(deepcopy, self.items))
    
    def map(self, fn):
        return BitMapDispatch(
            self.bitmap,
            [fn(elem) for elem in self.items]
        )


class DispatchNode(Node):
    """ Dispatch to children nodes depending of the hsh value at the
    current level. """
    __slots__ = ['children']
    def __init__(self, children=None):
        if children is None:
            children = BitMapDispatch()
        
        self.children = children
    
    def xor(self, hsh, shift, node):
        rlv = relevant(hsh, shift)
        newchild = self.children.get(rlv, NULLNODE).xor(hsh, shift + SHIFT, node)
        if newchild is NULLNODE:
            # This makes sure no dead nodes remain in the tree after
            # removing an item.
            newchildren = self.children.remove(rlv)
            if not newchildren:
                return NULLNODE
        else:
            newchildren = self.children.replace(
                rlv, 
                newchild
            )
        
        return DispatchNode(newchildren)
    
    def _ixor(self, hsh, shift, node):
        rlv = relevant(hsh, shift)
        newchild = self.children[rlv].xor(hsh, shift + SHIFT, node)
        if newchild is NULLNODE:
            self.children = self.children._iremove(rlv)
            if not self.children:
                return NULLNODE
        else:
            self.children = self.children._ireplace(rlv, newchild)
        
        return self
    
    @doc(ASSOC)
    def assoc(self, hsh, shift, node):
        # We need not check whether the return value of
        # self.children.get(...).assoc is NULLNODE, because assoc never
        # returns NULLNODE.
        rlv = relevant(hsh, shift)
        return DispatchNode(
            self.children.replace(
                rlv,
                self.children.get(rlv, NULLNODE).assoc(
                    hsh, shift + SHIFT, node
                )
            )
        )
    
    @doc(IASSOC)
    def _iassoc(self, hsh, shift, node):
        rlv = relevant(hsh, shift)
        self.children = self.children._ireplace(
            rlv, 
            self.children.get(rlv, NULLNODE)._iassoc(hsh, shift + SHIFT, node)
        )
        return self
    
    @classmethod
    def make(cls, shift, many):
        # Because the object we create in this function is not yet exposed
        # to any other code, we may safely call _iassoc.
        dsp = cls()
        for elem in many:
            dsp = dsp._iassoc(elem.hsh, shift, elem)
        return dsp
    
    @doc(GET)
    def get(self, hsh, shift, key):
        return self.children.get(relevant(hsh, shift), NULLNODE).get(
            hsh, shift + SHIFT, key
        )
    
    @doc(WITHOUT)
    def without(self, hsh, shift, key):
        rlv = relevant(hsh, shift)
        newchild = self.children[rlv].without(hsh, shift + SHIFT, key)
        if newchild is NULLNODE:
            # This makes sure no dead nodes remain in the tree after
            # removing an item.
            newchildren = self.children.remove(rlv)
            if not newchildren:
                return NULLNODE
        else:
            newchildren = self.children.replace(
                rlv, 
                newchild
            )
        
        return DispatchNode(newchildren)
    
    @doc(IWITHOUT)
    def _iwithout(self, hsh, shift, key):
        rlv = relevant(hsh, shift)
        newchild = self.children[rlv]._iwithout(hsh, shift + SHIFT, key)
        if newchild is NULLNODE:
            self.children = self.children._iremove(rlv)
            if not self.children:
                return NULLNODE
        else:
            self.children = self.children._ireplace(rlv, newchild)
        
        return self
    
    def __iter__(self):
        for child in self.children:
            for elem in child:
                yield elem
    
    def __copy__(self):
        return DispatchNode(self.children.map(copy))



from burrahobbit._tree import BitMapDispatch, ListDispatch

def test_dispatch():
    nd = BitMapDispatch()
    for key in xrange(16):
        nd = nd.replace(key, None)
    assert isinstance(nd, BitMapDispatch)
    nd = nd.replace(17, None)
    assert isinstance(nd, ListDispatch)



# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import pytest

from burrahobbit import set as bset
from burrahobbit.treeset import PersistentTreeSet


class HashCollision(object):
    def __init__(self, item, hsh):
        self.item = item
        self.hsh = hsh
    
    def __hash__(self):
        return self.hsh
    
    def __eq__(self, other):
        return isinstance(other, HashCollision) and self.item == other.item


def random_set(size):
    return set(os.urandom(20) for _ in xrange(size))


def test_or():
    some = random_set(1000)
    some.update(['a', 'b', 'c'])
    other = random_set(1000)
    other.update(set(['a', 'c', 'd']))
    
    df = PersistentTreeSet.from_set(some) | PersistentTreeSet.from_set(other)
    some.update(other)
    assert set(df) == some


def test_xor():
    some = random_set(1000)
    some.update(['a', 'b', 'c'])
    other = random_set(1000)
    other.update(set(['a', 'c', 'd']))
    
    df = PersistentTreeSet.from_set(some) ^ PersistentTreeSet.from_set(other)
    
    assert set(df) == some ^ other


def test_and():
    some = random_set(1000)
    some.update(['a', 'b', 'c'])
    other = random_set(1000)
    other.update(set(['a', 'c', 'd']))
    
    df = PersistentTreeSet.from_set(some) & PersistentTreeSet.from_set(other)
    
    assert set(df) == some & other


def test_fromset():
    st = random_set(1000)
    mp = PersistentTreeSet.from_set(st)
    for elem in st:
        assert elem in mp


def test_persistence():
    mp = PersistentTreeSet()
    mp1 = mp.add('a')
    assert 'a' in mp1
    mp2 = mp1.add('b')
    assert 'a' in mp2
    assert 'b' in mp2
    assert 'b' not in mp1
    mp3 = mp2.without('a')
    assert 'b' in mp3
    assert 'a' not in mp3


def test_iteration():
    st = random_set(1000)
    mp = PersistentTreeSet.from_set(st)
    assert set(mp) == st


def test_transient():
    mp = PersistentTreeSet.from_set(set(['foo']))
    mp2 = mp.transient()
    mp3 = mp2.add('bar')
    assert mp2 is mp3
    assert 'foo' in mp2
    assert 'bar' in mp2
    assert 'foo' in mp3
    assert 'bar' in mp3
    
    mp4 = mp3.persistent()
    mp5 = mp4.add('baz')
    assert 'baz' not in mp2
    assert 'baz' not in mp3
    assert 'baz' not in mp4
    assert 'baz' in mp5
    assert mp5 is not mp4


def test_collision():
    HASH = 13465345
    mp = PersistentTreeSet()
    mp = mp.add(HashCollision("hello", HASH))
    mp = mp.add(HashCollision("answer", HASH))
    assert HashCollision("hello", HASH) in mp
    assert HashCollision("answer", HASH) in mp


def test_eq():
    assert bset(['foo', 'bar']) == bset(['bar', 'foo'])


def test_construct():
    assert (
        bset(set(['foo', 'bar'])) ==
        bset(['foo', 'bar']) ==
        bset({'foo': 1, 'bar': 2}) ==
        bset(iter(['foo', 'bar'])) ==
        bset(bset(['foo', 'bar']))
    )



# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import pytest

from copy import copy

from burrahobbit import dict as bdict
from burrahobbit.treedict import PersistentTreeMap


class HashCollision(object):
    def __init__(self, item, hsh):
        self.item = item
        self.hsh = hsh
    
    def __hash__(self):
        return self.hsh
    
    def __eq__(self, other):
        return isinstance(other, HashCollision) and self.item == other.item


def random_dict(size):
    return dict((os.urandom(20), os.urandom(25)) for _ in xrange(size))


def test_or():
    some = random_dict(1000)
    some.update({'a': 'foo', 'b': 'bar', 'c': 'blub'})
    other = random_dict(1000)
    other = {'a': 'blub', 'c': 'blab', 'd': 'quuz'}
    
    df = PersistentTreeMap.from_dict(some) | PersistentTreeMap.from_dict(other)
    some.update(other)
    assert set(df.iteritems()) == set(some.iteritems())


def test_xor():
    some = random_dict(1000)
    some.update({'a': 'foo', 'b': 'bar', 'c': 'blub'})
    other = random_dict(1000)
    other.update({'a': 'blub', 'c': 'blab', 'd': 'quuz'})
    
    df = PersistentTreeMap.from_dict(some) ^ PersistentTreeMap.from_dict(other)
    
    for key, value in df.iteritems():
        assert (
            key in some and key not in other and some[key] == value
        ) or key in other and key not in some and other[key] == value


def test_and():
    some = random_dict(1000)
    some.update({'a': 'foo', 'b': 'bar', 'c': 'blub'})
    other = random_dict(1000)
    other.update({'a': 'blub', 'c': 'blab', 'd': 'quuz'})
    
    third = copy(some)
    third.update(other)
    
    df = PersistentTreeMap.from_dict(some) & PersistentTreeMap.from_dict(other)
    
    for key in third:
        if key in some and key in other:
            assert df[key] == other[key]


def test_fromdict():
    dct = random_dict(1000)
    mp = PersistentTreeMap.from_dict(dct)
    for key, value in dct.iteritems():
        assert mp[key] == value


def test_persistence():
    mp = PersistentTreeMap()
    mp1 = mp.assoc('a', 'hello')
    assert mp1['a'] == 'hello'
    mp2 = mp1.assoc('b', 'world')
    assert mp2['a'] == 'hello'
    assert mp2['b'] == 'world'
    mp3 = mp2.without('a')
    assert mp3['b'] == 'world'
    with pytest.raises(KeyError) as excinfo:
        assert mp3['a'] == 'hello'
    assert excinfo.value.args[0] == 'a'


def test_iteration():
    dct = random_dict(1000)
    mp = PersistentTreeMap.from_dict(dct)
    assert set(mp.iterkeys()) == set(mp) == set(dct.iterkeys())
    assert set(mp.itervalues()) == set(dct.itervalues())
    assert set(mp.iteritems()) == set(dct.iteritems())


def test_novaluecopy():
    mp = PersistentTreeMap()
    bar = []
    mp = mp.assoc("foo", bar)
    mv = mp.transient()
    bar.append("test")
    assert mp["foo"] == mv["foo"] == bar


def test_transient():
    mp = PersistentTreeMap.from_dict({'foo': 'baz'})
    mp2 = mp.transient()
    mp3 = mp2.assoc('foo', 'bar')
    assert mp2['foo'] == 'bar'
    assert mp3['foo'] == 'bar'
    assert mp2 is mp3
    assert mp['foo'] == 'baz'
    
    mp4 = mp3.persistent()
    mp5 = mp4.assoc('foo', 'spam')
    assert mp2['foo'] == 'bar'
    assert mp3['foo'] == 'bar'
    assert mp4['foo'] == 'bar'
    assert mp5['foo'] == 'spam'


def test_collision():
    HASH = 13465345
    mp = PersistentTreeMap()
    mp = mp.assoc(HashCollision("hello", HASH), "world")
    mp = mp.assoc(HashCollision("answer", HASH), 42)
    assert mp[HashCollision("hello", HASH)] == "world"
    assert mp[HashCollision("answer", HASH)] == 42


def test_neq():
    some = random_dict(1000)
    some.update({'a': 'foo', 'b': 'bar', 'c': 'blub'})
    other = random_dict(1000)
    other.update({'a': 'blub', 'c': 'blab', 'd': 'quuz'})
    assert (
        PersistentTreeMap.from_dict(some) != PersistentTreeMap.from_dict(other)
    )


def test_eq():
    some = random_dict(1000)
    assert (
        PersistentTreeMap.from_dict(some) == PersistentTreeMap.from_dict(some)
    )
    assert (
        PersistentTreeMap.from_dict(some) ==
        PersistentTreeMap.from_itr(sorted(some.iteritems()))
    )


def test_construct():
    assert (
        bdict(foo=1, bar=2) ==
        bdict({'foo': 1, 'bar': 2}) ==
        bdict([('foo', 1), ('bar', 2)]) ==
        bdict(iter([('foo', 1), ('bar', 2)])) ==
        bdict(bdict(foo=1, bar=2))
    )


def main():
    import os
    import time
    
    mp = PersistentTreeMap().transient()
    for _ in xrange(22500):
        one, other = os.urandom(20), os.urandom(25)
        mp2 = mp.assoc(one, other)
        assert mp[one] == other
        assert mp2[one] == other
        mp = mp2
    pmp = mp.persistent()    
    
    s = time.time()
    mp = PersistentTreeMap()
    for _ in xrange(225000):
        one, other = os.urandom(20), os.urandom(25)
        mp2 = mp.assoc(one, other)
        try:
            mp[one]
        except KeyError:
            assert True
        else:
            assert False
        try:
            mp2.without(one)[one]
        except KeyError:
            assert True
        else:
            assert False
        mp = mp2
        assert mp[one] == other
    print 'PersistentHashMap:', time.time() - s
    assert mp[one] == other
    # This /may/ actually fail if we are unlucky, but it's a good start.
    assert len(list(iter(mp))) == 225000


if __name__ == '__main__':
    main()



# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from copy import copy

from burrahobbit._tree import (
    NULLNODE, GET, ASSOC, IASSOC, WITHOUT, doc, DispatchNode, HashCollisionNode
)

class SetNode(object):
    """ A AssocNode contains the actual key-value mapping. """
    __slots__ = ['key', 'hsh']
    def __init__(self, key):
        self.key = key
        self.hsh = hash(key)
    
    def xor(self, hsh, shift, node):
        if node.key == self.key:
            return NULLNODE
        else:
            return self.assoc(hsh, shift, node)
    
    @doc(GET)
    def get(self, hsh, shift, key):
        # If the key does not match the key of the AssocNode, thus the hash
        # matches to the current level, but it is not the correct node,
        # raise a KeyError, otherwise return the value.
        if key != self.key:
            raise KeyError(key)
        return self
    
    @doc(ASSOC)
    def assoc(self, hsh, shift, node):
        # If there is a hash-collision, return a HashCollisionNode,
        # otherwise return a DispatchNode dispatching depending on the
        # current level (if the two hashes only differ at a higher-level,
        # DispatchNode.make will return a DispatchNode that contains a
        # DispatchNode etc. up until the necessary depth.
        if node.key == self.key:
            return node
        
        if hsh == self.hsh:
            return HashCollisionNode(
                [self, node]
            )
        return DispatchNode.make(shift, [self, node])
    
    @doc(IASSOC)
    def _iassoc(self, hsh, shift, node):
        if node.key == self.key:
            return node
        
        if hsh == self.hsh:
            return HashCollisionNode(
                [self, node]
            )
        return DispatchNode.make(shift, [self, node])        
    
    @doc(WITHOUT)
    def without(self, hsh, shift, key):
        # If the key matches the key of this AssocNode, returning NULLNODE
        # will remove the Node from the map. Otherwise raise a KeyError.
        if key != self.key:
            raise KeyError(key)
        return NULLNODE
    
    _iwithout = without
    
    def __iter__(self):
        yield self
    
    def __copy__(self):
        return SetNode(self.key)
    
    def __eq__(self, other):
        return self.key == other.key
    
    def __neq__(self, other):
        return self.key != other.key


class PersistentTreeSet(object):
    __slots__ = ['root']
    def __init__(self, root=NULLNODE):
        self.root = root
    
    def __contains__(self, key):
        try:
            self.root.get(hash(key), 0, key)
            return True
        except KeyError:
            return False
    
    def __and__(self, other):
        return other.__class__(self.root & other.root)
    
    def __xor__(self, other):
        return PersistentTreeSet(self.root ^ other.root)
    
    def __or__(self, other):
        return PersistentTreeSet(self.root | other.root)
    
    def __eq__(self, other):
        return self.root == other.root
    
    def __neq__(self, other):
        return self.root != other.root
    
    def add(self, key):
        """ Return copy of self with an association between key and value.
        May override an existing association. """
        return PersistentTreeSet(
            self.root.assoc(hash(key), 0, SetNode(key))
        )
    
    def without(self, key):
        """ Return copy of self with key removed. """
        return PersistentTreeSet(
            self.root.without(hash(key), 0, key)
        )
    
    def __iter__(self):
        for node in self.root:
            yield node.key
    
    @staticmethod
    def from_set(set_):
        """ Create PersistentTreeSet from existing set. """
        mp = TransientTreeSet()
        for key in set_:
            mp = mp.add(key)
        return mp.persistent()
    
    @staticmethod
    def construct(iterable=None):
        if isinstance(iterable, PersistentTreeSet):
            return PersistentTreeSet(copy(iterable.root))
        if iterable is None:
            return PersistentTreeSet()
        
        return PersistentTreeSet.from_set(iterable)
        
    def transient(self):
        """ Return transient (mutable) copy of self. Changing the copy will not
        affect the original object's immutability.
        
        See :class:`TransientTreeSet`. """
        return TransientTreeSet(copy(self.root))


class TransientTreeSet(PersistentTreeSet):
    def add(self, key):
        """ Update this TransientTreeMap to contain an association between
        key and value.
        
        USE WITH CAUTION: This should only be used if no other reference
        to the PersistentTreeMap may exist. """
        self.root = self.root.assoc(hash(key), 0, SetNode(key))
        return self
    
    def without(self, key):
        """ Remove key.
        
        USE WITH CAUTION: This should only be used if no other reference
        to the PersistentTreeMap may exist. """
        self.root = self.root._iwithout(hash(key), 0, key)
        return self
    
    def persistent(self):
        """ Return a persistent version of self.
        
        CAUTION: The :class:`TransientTreeMap` MAY NOT BE USED
        after calling this method.
        """
        return PersistentTreeSet(self.root)



#!/usr/bin/env python
# -*- coding: us-ascii -*-

# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys

try:
    # If we got setuptools, use it so we get the nice develop cmd.
    from setuptools import setup
except:
    # Doesn't matter if we don't have it though.
    from distutils.core import setup


VERSION = '0.1.1'

extra = {}
if sys.version_info >= (3, 0):
    extra['use_2to3'] = True

setup(
    name='burrahobbit',
    version=VERSION,
    description='Persistent data-structures in Python.',
    author='Florian Mayer',
    author_email='florian.mayer@bitsrc.org',
    url='http://segfaulthunter.github.com/burrahobbit',
    keywords='persistent datastructures',
    license='MIT',
    zip_safe=True,
    packages=['burrahobbit', 'burrahobbit.test'],
    **extra
)



# Copyright (C) 2011 by Florian Mayer <florian.mayer@bitsrc.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from copy import copy

from burrahobbit._tree import NULLNODE, SENTINEL
from burrahobbit.treeset import SetNode

class AssocNode(SetNode):
    """ A AssocNode contains the actual key-value mapping. """
    __slots__ = ['value']
    def __init__(self, key, value):
        SetNode.__init__(self, key)
        self.value = value
    
    def __repr__(self):
        return '<AssocNode(%r, %r)>' % (self.key, self.value)
    
    def __copy__(self):
        return AssocNode(self.key, self.value)
    
    def __eq__(self, other):
        return self.key == other.key and self.value == other.value
    
    def __neq__(self, other):
        return self.key != other.key or self.value != other.value


class PersistentTreeMap(object):
    __slots__ = ['root']
    def __init__(self, root=NULLNODE):
        self.root = root
    
    def __getitem__(self, key):
        return self.root.get(hash(key), 0, key).value
    
    def __and__(self, other):
        return other.__class__(self.root & other.root)
    
    def __xor__(self, other):
        return PersistentTreeMap(self.root ^ other.root)
    
    def __or__(self, other):
        return PersistentTreeMap(self.root | other.root)
    
    def __eq__(self, other):
        return self.root == other.root
    
    def __neq__(self, other):
        return self.root != other.root
    
    def assoc(self, key, value):
        """ Return copy of self with an association between key and value.
        May override an existing association. """
        return PersistentTreeMap(
            self.root.assoc(hash(key), 0, AssocNode(key, value))
        )
    
    def without(self, key):
        """ Return copy of self with key removed. """
        return PersistentTreeMap(
            self.root.without(hash(key), 0, key)
        )
    
    def __iter__(self):
        """ Yield keys for all items. """
        for node in self.root:
            yield node.key
    
    iterkeys = __iter__
    
    def iteritems(self):
        """ Yield key, value pairs for all items. """
        for node in self.root:
            yield node.key, node.value
    
    def itervalues(self):
        """ Yield values for all items. """
        for node in self.root:
            yield node.value
    
    @staticmethod
    def from_itr(itr):
        mp = TransientTreeMap()
        for key, value in itr:
            mp = mp.assoc(key, value)
        return mp.persistent()        
    
    @staticmethod
    def from_dict(dct):
        """ Create PersistentTreeMap from existing dictionary. """
        mp = TransientTreeMap()
        for key, value in dct.iteritems():
            mp = mp.assoc(key, value)
        return mp.persistent()
    
    def transient(self):
        """ Return transient (mutable) copy of self. Changing the copy will not
        affect the original object's immutability.
        
        See :class:`TransientTreeMap`. """
        return TransientTreeMap(copy(self.root))
    
    @staticmethod
    def construct(argument=SENTINEL, **kwargs):
        if kwargs:
            if argument is not SENTINEL:
                kwargs['argument'] = argument
            return PersistentTreeMap.from_dict(kwargs)
        
        if argument is SENTINEL:
            return PersistentTreeMap()
        
        if isinstance(argument, dict):
            return PersistentTreeMap.from_dict(argument)
        
        if isinstance(argument, TransientTreeMap):
            return PersistentTreeMap(copy(argument.root))
        
        if isinstance(argument, PersistentTreeMap):
            return argument
                
        return PersistentTreeMap.from_itr(argument)


class TransientTreeMap(PersistentTreeMap):
    """
    TransientTreeMaps are used if - and only if - one function does so many
    changes to a PersistentTreeMap the immutability would prove inefficient.
    
    The function has to return transienttreemap.persistent() in order to ensure
    that the treemap cannot be changed afterwards. """
    def assoc(self, key, value):
        """ Update this TransientTreeMap to contain an association between
        key and value and return self. You should never assume that the
        object really is mutated as this only is an optimization, thus, you
        should always bind the return value of this function to the 
        respective name, e.g., `mymap.assoc("spam", "eggs")` should be avoided
        and written as `mymap = mymap.assoc("spam", "eggs")` instead. """
        self.root = self.root._iassoc(hash(key), 0, AssocNode(key, value))
        return self
    
    def without(self, key):
        """ Remove key. """
        self.root = self.root._iwithout(hash(key), 0, key)
        return self
    
    def persistent(self):
        """ Return a persistent version of self.
        
        CAUTION: The :class:`TransientTreeMap` MAY NOT BE USED
        after calling this method.
        """
        return PersistentTreeMap(self.root)
